#ifndef carbonquant_Calibration_h
#define carbonquant_Calibration_h

#include <carbonquantCalibrationModelSpotExport.h>

#include <string>
#include <vector>
#include <map>

class Spot;
class Optimizer;

class CARBONQUANTCALIBRATIONMODELSPOT_EXPORT Calibration
{
protected:
  std::vector<double> spot_obs; // Spot values
  std::vector<double> rentability_obs; // Observed value that will be calibrated initialized by normalization of spot_obs

  // Initial parameters of model
  std::vector<double> model_param;

  // Best parameters of model found
  std::vector<double> best_model_param;

  std::vector<double> lower_bound_model_param;
  std::vector<double> upper_bound_model_param;

  double time_step;

  double spot_scale; // parameter s of :: S_t = s ( exp(X_t) - a )

  double const_a; // parameter a of :: S_t = s ( exp(X_t) - a )

public: 
  Calibration(double const_a = 0);
  virtual ~Calibration();

  typedef std::map<int, std::string> mapIntString;

  virtual void init();

  // Init parameters from spot
  virtual void initParametersFromSpot(const Spot* spot, const std::map<std::string, double>& spot_bound, const std::vector<double>& spotObs, const double& timeStep, const double& spotScale);

  // Set spot from best model parameters
  virtual void initSpotFromParameters(Spot *spot);

  // Optimizer
  void setOptimizer(Optimizer* opt);

  void setLowerBoundModelParam(const std::vector<double>& lower_bound);
  void setUpperBoundModelParam(const std::vector<double>& upper_bound);

  bool check_point(const std::vector<double>& point) const;

  // function to compute the objective function
  virtual std::vector<double> f(const std::vector<double>& point) const = 0;

  // CDF = cumulative density function (in french : fonction de repartition) 
  // observations
  std::vector<double> cdf_obs(const double& born_min, const double& born_max, const int& size_of_cdf_obs, const std::vector<double>& deltaRentabilityObs) const;
  // calibrated model
  virtual std::vector<double> cdf_model(const double& born_min, const double& born_max, const int& size_of_cdf_obs) const;
  virtual std::vector<double> cdf_model_day_night(const double& born_min, const double& born_max, const int& size_of_cdf_obs, const std::vector<double>& model_param_day, const std::vector<double>& model_param_night) const;

  std::vector<double> delta_rentability_obs() const;
  virtual double delta_rentability_obs(const double& rentabilityObs, const double& rentabilityObsAfter, const std::vector<double>& model_param) const = 0;

  std::vector<double> delta_rentability_obs_day_night(const std::vector<double>& rentability_obs_day_night, const std::vector<double>& model_param_day, const std::vector<double>& model_param_night, const mapIntString& indicator_day_night) const;
  std::vector<double> rentability_model() const;

  virtual double rentability_model(const double& rentabilityModelBefore, const std::vector<double>& model_param) const = 0;

  std::vector<double> rentability_model_day_night(const std::vector<double>& rentability_obs_day_night, const std::vector<double>& model_param_day, const std::vector<double>& model_param_night, const mapIntString& indicator_day_night) const;

  // Compute calibration
  virtual void compute(Optimizer* solver);

  // Normalize spot values of observable
  // Change spot_obs in rentability_obs
  void normalize_spot_obs();
  std::vector<double> compute_rentability_obs(const double& scale);

  // Get model param
  std::vector<double> get_model_param() const;

  // Get best model param
  std::vector<double> get_best_model_param() const;

  // Get spot scale
  inline double getSpotScale() const {return spot_scale;}; 

  virtual std::string name_of_model() const = 0;

  std::vector<std::vector<double> > trajectory_data(std::vector<double> model_param_day = std::vector<double>(), std::vector<double> model_param_night = std::vector<double>(), mapIntString indicator_day_night = mapIntString(), double spot_scale_mean_day_night = 0);

  std::vector<std::vector<double> > cdf_data(std::vector<double> model_param_day = std::vector<double>(), std::vector<double> model_param_night = std::vector<double>(), mapIntString indicator_day_night = mapIntString(), double spot_scale_mean_day_night = 0);

  //--------------------------------------
  // Function to create output file
  //--------------------------------------
  // This file contains comparison of:
  //           - obs and a trajectory of calibrated model
  //           - CDF of obs and CDF of calibrated model
  void create_data(std::string outputDir = "./", std::vector<double> model_param_day = std::vector<double>(), std::vector<double> model_param_night = std::vector<double>(), mapIntString indicator_day_night = mapIntString(), double spot_scale_mean_day_night = 0); // function to create output file

  static void set_spot_obs_day_night(const std::vector<double>& spot_values, const std::vector<int>& dayNight, const int& firstHour, std::vector<double>& spot_obs_day, std::vector<double>& spot_obs_night, mapIntString& indicator_day_night, double& spot_scale_mean_day_night);

  static double compute_spot_scale(const std::vector<double>& spot_values, const int& choice_of_spot_scale);

protected:
  double pdf_normal(double,double=0,double=1) const;
  double moment_gamma(double,double,double) const;
  double moment_normal(int) const;
  double cdf_nor(double) const;
  double gamma_fct(double x) const;
  double Cnp(int n, int p) const;
};

#endif // carbonquant_Calibration
 
