#ifndef carbonquant_Spot_h
#define carbonquant_Spot_h

#include "carbonquantModelSpotExport.h"

#include <cmath>
#include <iostream>
#include <assert.h>
#include <stdlib.h>
#include <functional>
#include <algorithm>
#include <numeric>
#include <vector>
#include <math.h>

/** A simple Electricity Spot Class
*/
class CARBONQUANTMODELSPOT_EXPORT Spot 
{
public:
    /* Model description
        BM           => Brownian motion
        BLACKSCHOLES => Black Scholes
        OU           => Processus Ornstein-Uhlebeck
        MODIFIED_OU  => "Modified" processus Ornstein-Uhlebeck
        EXP_OU       => Exponential of Ornstein-Uhlebeck
        EXP_OU2      => Exponential of (Ornstein-Uhlebeck)^2/2
        EXP_CIR      => Exponential of Cox Ingersoll Ross
        EXP_OU_VG    => Exponential of Ornstein-Uhlebeck with Variance Gamma motion
    */
    enum Model
    {
        BM=0,
        BLACKSCHOLES=1,
        OU=2,
        MODIFIED_OU=3,
        EXP_OU=4,
        EXP_OU2=5,
        EXP_CIR=6,
        EXP_OU_VG=7
    };

protected:
    double m_time;
    std::vector<double> m_state;

    std::vector<std::vector<double> > m_diffusion;
    std::vector<double> m_drift;    
  
public:
             Spot(void);
    virtual ~Spot(void);

// Copy
public:
    Spot(const Spot& spot);
    Spot& operator = (const Spot& spot);

public:
    virtual Spot* clone(void) const;

    //----------- returns kind of the spot -------------------------------------
    virtual std::string identifier(void) const { return std::string(); };

    virtual Model get_model(void) const {return OU;};

    void setTime(double time) {m_time = time;};
    virtual void setState(const std::vector<double>& state);

    //
    virtual const std::vector<std::vector<double> >& diffusion(void) { return m_diffusion; };
    virtual const std::vector<double>& drift(void) { return m_drift; };

    //--------------------------------------------------------------------------
    virtual void set_volatility(const double&) {};
    virtual void set_meanParameter(const double&) {}; 
    virtual void set_meanReversion(const double&) {};
    virtual void set_coeffDrift(const double&) {};
    virtual void set_rentability(const double&) {};
    virtual void set_vgGammaShape_kappa(const double&) {};
    virtual void set_vgBrownianDriftCoef_theta(const double&) {};
    virtual void set_vgAdditionalDrift_m(const double&) {};
    virtual void set_spot_scale(const double&) {}; 

    //--------------------------------------------------------------------------
    virtual double get_volatility() const { return 0.; };
    virtual double get_meanParameter() const { return 0.; }; 
    virtual double get_meanReversion() const { return 0.; };
    virtual double get_coeffDrift() const { return 0.; };
    virtual double get_rentability() const { return 0.; };
    virtual double get_vgGammaShape_kappa() const { return 0.; };
    virtual double get_vgBrownianDriftCoef_theta() const { return  0.; };
    virtual double get_vgAdditionalDrift_m() const { return 0.; };
    virtual double get_spot_scale() const { return 0.; };

    //--------------------------------------------------------------------------
    virtual void set_spot_day(Spot *spot_day) {};
    virtual void set_spot_night(Spot *spot_night) {};

    virtual void set_begin_day(const int& begin_day) {};
    virtual void set_begin_night(const int& begin_night) {};
 
    virtual const Spot* get_spot_day() const {return NULL;};
    virtual const Spot* get_spot_night() const {return NULL;};

    //-----------  function to simulate a spot ----------------------------------
    void set_begin_hour(double begin) {};
    void set_scale_hour(double scale) {};

    virtual double simulate(const double &dt, const double &spotValue) const {return 0;};
    virtual void simulate(const double &dt, std::vector<double>& spotData) const;

    virtual void print_model_param() const {};

    // Read spot price from file
    static double* readFile(const int& nInt, const std::string& filename);
};

typedef Spot::Model SpotModel;

#endif // carbonquant_Spot_h
